"""
Abstract base classes defining Estimators of individual galaxy redshift uncertainties.
"""

import gc
import sys
from typing import Any, Optional

import numpy as np
import qp

from rail.core.common_params import SHARED_PARAMS
from rail.core.data import (DataHandle, ModelHandle, ModelLike, QPHandle,
                            TableHandle, TableLike)
from rail.core.point_estimation import PointEstimationMixin
from rail.core.stage import RailStage
from rail.core.enums import DistributionType

# for backwards compatibility, to avoid break stuff that imports it from here
from .informer import CatInformer  # pylint: disable=unused-import
import tables_io


class CatEstimator(RailStage, PointEstimationMixin):
    """The base class for making photo-z posterior estimates from catalog-like inputs
    (i.e., tables with fluxes in photometric bands among the set of columns)

    Estimators use a generic "model", the details of which depends on the sub-class.

    Estimators take as "input" tabular data, apply the photo-z estimation and
    provide as "output" a ``QPEnsemble``, with per-object p(z).

    """

    name = "CatEstimator"
    config_options = RailStage.config_options.copy()
    config_options.update(
        chunk_size=SHARED_PARAMS,
        hdf5_groupname=SHARED_PARAMS,
        zmin=SHARED_PARAMS,
        zmax=SHARED_PARAMS,
        nzbins=SHARED_PARAMS,
        id_col=SHARED_PARAMS,
        redshift_col=SHARED_PARAMS,
        calc_summary_stats=SHARED_PARAMS,
    )
    config_options.update(
        **PointEstimationMixin.config_options.copy(),
    )
    inputs = [("model", ModelHandle), ("input", TableHandle)]
    outputs = [("output", QPHandle)]

    def __init__(self, args: Any, **kwargs: Any) -> None:
        """Initialize Estimator"""
        super().__init__(args, **kwargs)
        self._output_handle: QPHandle | None = None
        self.model = None

    def estimate(self, input_data: TableLike) -> DataHandle:
        """The main interface method for the photo-z estimation

        This will attach the input data (defined in ``inputs`` as "input") to this
        ``Estimator`` (for introspection and provenance tracking). Then call the
        ``run()``, ``validate()``, and ``finalize()`` methods.

        The run method will call ``_process_chunk()``, which needs to be implemented
        in the subclass, to process input data in batches. See ``RandomGaussEstimator``
        for a simple example.

        Finally, this will return a ``QPHandle`` for access to that output data.

        Parameters
        ----------
        input_data
            A dictionary of all input data

        Returns
        -------
        DataHandle
            Handle providing access to QP ensemble with output data
        """
        self.set_data("input", input_data)
        self.validate()
        self.run()
        self.finalize()
        results = self.get_handle("output")
        results.read(force=True)
        return results

    def run(self) -> None:
        self.open_model(**self.config)

        iterator = self.input_iterator("input")
        first = True
        self._initialize_run()
        self._output_handle = None
        for s, e, test_data in iterator:
            print(f"Process {self.rank} running estimator on chunk {s:,} - {e:,}")
            self._process_chunk(s, e, test_data, first)
            first = False
            # Running garbage collection manually seems to be needed
            # to avoid memory growth for some estimators
            gc.collect()
        self._finalize_run()

    def _initialize_run(self) -> None:
        self._output_handle = None

    def _finalize_run(self) -> None:
        assert self._output_handle is not None
        self._output_handle.finalize_write()

    def _process_chunk(
        self, start: int, end: int, data: TableLike, first: bool
    ) -> None:
        raise NotImplementedError(
            f"{self.name}._process_chunk is not implemented"
        )  # pragma: no cover

    @classmethod
    def default_distribution_type(cls) -> DistributionType:
        """Return the type of distribtuion that this estimator creates

        By default this is DistributionType.ad_hoc
        But this can be overrided by sub-classes to return
        DistributionType.posetrior or DistributionType.likelihood if appropriate
        """
        return DistributionType.ad_hoc

    def _calculate_summary_stats(
        self,
        qp_dstn: qp.Ensemble,
    ) -> qp.Ensemble:

        if qp_dstn.ancil is None:  # pragma: no cover
            ancil_dict: dict[str, np.ndarray] = dict()
            qp_dstn.set_ancil(ancil_dict)

        quantiles = [0.025, 0.16, 0.5, 0.85, 0.975]
        quant_names = ['q2p5', 'q16', 'median', 'q84', '97p5']

        locs = qp_dstn.ppf(quantiles)
        for name_, vals_ in zip(quant_names, locs.T):
            qp_dstn.ancil[f"z_{name_}"] = np.expand_dims(vals_, -1)

        grid: np.ndarray | None = None

        if 'z_mode' not in qp_dstn.ancil:
            grid = np.linspace(self.config.zmin, self.config.zmax, self.config.nzbins)
            qp_dstn.ancil['z_mode'] = qp_dstn.mode(grid)

        try:
            qp_dstn.ancil['z_mean'] = qp_dstn.mean()
            qp_dstn.ancil['z_std'] = qp_dstn.std()
        except IndexError:  # pragma: no cover
            # this is needed b/c qp.MixMod pdf sometimes fails to compute moments
            grid = np.linspace(self.config.zmin, self.config.zmax, self.config.nzbins)
            pdfs = qp_dstn.pdf(grid)
            norms = pdfs.sum(axis=1)
            means = np.sum(pdfs * grid, axis=1) / norms
            diffs = (np.expand_dims(grid, -1) - means).T
            wt_diffs = diffs * diffs * pdfs
            stds = np.sqrt((wt_diffs).sum(axis=1)/norms)
            qp_dstn.ancil['z_mean'] = np.expand_dims(means, -1)
            qp_dstn.ancil['z_std'] = np.expand_dims(stds, -1)

        return qp_dstn

    def _do_chunk_output(
        self,
        qp_dstn: qp.Ensemble,
        start: int,
        end: int,
        first: bool,
        data: Optional[TableLike] = None,
    ) -> None:
        qp_dstn = self.calculate_point_estimates(qp_dstn)

        if self.config.calc_summary_stats:
            qp_dstn = self._calculate_summary_stats(qp_dstn)

        # if there is no ancil set by the calculate_point_estimate, initiate one
        if data is not None:
            if qp_dstn.ancil is None:  # pragma: no cover
                ancil_dict: dict[str, np.ndarray] = dict()
                qp_dstn.set_ancil(ancil_dict)
            assert qp_dstn.ancil is not None
            # if there is ID column in the input dataset, attach it to the ancil
            if self.config.id_col in data.keys():  # pragma: no cover
                qp_dstn.ancil.update(id=data[self.config.id_col])
            # if there is redshift column in the input dataset, attach it to the ancil
            if self.config.redshift_col in data.keys():  # pragma: no cover
                qp_dstn.ancil.update(redshift=data[self.config.redshift_col])

            if 'distribution_type' not in qp_dstn.ancil:
                qp_dstn.ancil.update(distribution_type=np.repeat(self.default_distribution_type().value, end-start))

        if first:
            the_handle = self.add_handle("output", data=qp_dstn)
            assert isinstance(the_handle, QPHandle)
            self._output_handle = the_handle
            if self.config.output_mode != "return":
                self._output_handle.initialize_write(
                    self._input_length, communicator=self.comm
                )
        assert self._output_handle is not None
        self._output_handle.set_data(qp_dstn, partial=True)
        if self.config.output_mode != "return":
            self._output_handle.write_chunk(start, end)
        return qp_dstn


    def _convert_table_format(self, data: TableLike, out_fmt_str: str="numpyDict") -> TableLike: # pragma: no cover
        """
        Utility function to convert existing Tabular data to a numpy dictionary,
        ingestable for most informer and estimators.
        To be called in _process_chunk().
        """
        # required format for informer/estimator
        out_fmt = tables_io.types.TABULAR_FORMAT_NAMES[out_fmt_str] 
        out_data = tables_io.convert(data, out_fmt)
        # overwrite set_data
        return out_data


class PzEstimator(RailStage, PointEstimationMixin):
    """The base class for making photo-z posterior estimates from other pz inputs

    Estimators use a generic "model", the details of which depends on the sub-class.

    Estimators take as "input" a QPEnsemble, with other estimates and
    provide as "output" a ``QPEnsemble``, with per-object p(z).

    """

    name = "PzEstimator"
    config_options = RailStage.config_options.copy()
    config_options.update(
        chunk_size=SHARED_PARAMS,
        hdf5_groupname=SHARED_PARAMS,
    )
    config_options.update(
        **PointEstimationMixin.config_options.copy(),
    )
    inputs = [("model", ModelHandle), ("input", QPHandle)]
    outputs = [("output", QPHandle)]

    def __init__(self, args: Any, **kwargs: Any) -> None:
        """Initialize Estimator"""
        super().__init__(args, **kwargs)
        self._output_handle: QPHandle | None = None
        self.model = None

    def estimate(self, input_data: QPHandle) -> DataHandle:
        """The main interface method for the photo-z estimation

        This will attach the input data (defined in ``inputs`` as "input") to this
        ``Estimator`` (for introspection and provenance tracking). Then call the
        ``run()``, ``validate()``, and ``finalize()`` methods.

        The run method will call ``_process_chunk()``, which needs to be implemented
        in the subclass, to process input data in batches. See ``RandomGaussEstimator``
        for a simple example.

        Finally, this will return a ``QPHandle`` for access to that output data.

        Parameters
        ----------
        input_data
            A dictionary of all input data

        Returns
        -------
        DataHandle
            Handle providing access to QP ensemble with output data
        """
        self.set_data("input", input_data)
        self.validate()
        self.run()
        self.finalize()
        results = self.get_handle("output")
        results.read(force=True)
        return results

    def run(self) -> None:
        self.open_model(**self.config)

        iterator = self.input_iterator("input")
        first = True
        self._initialize_run()
        self._output_handle = None
        for s, e, test_data in iterator:
            print(f"Process {self.rank} running estimator on chunk {s:,} - {e:,}")
            sys.stdout.flush()
            self._process_chunk(s, e, test_data, first)
            first = False
            # Running garbage collection manually seems to be needed
            # to avoid memory growth for some estimators
            gc.collect()
        self._finalize_run()

    def _initialize_run(self) -> None:
        self._output_handle = None

    def _finalize_run(self) -> None:
        assert self._output_handle is not None
        self._output_handle.finalize_write()

    def _process_chunk(
        self, start: int, end: int, data: qp.Ensemble, first: bool
    ) -> None:
        raise NotImplementedError(
            f"{self.name}._process_chunk is not implemented"
        )  # pragma: no cover

    def _do_chunk_output(
        self, qp_dstn: qp.Ensemble, start: int, end: int, first: bool
    ) -> None:
        qp_dstn = self.calculate_point_estimates(qp_dstn)
        if first:
            the_handle = self.add_handle("output", data=qp_dstn)
            assert isinstance(the_handle, QPHandle)
            self._output_handle = the_handle
            if self.config.output_mode != "return":
                self._output_handle.initialize_write(
                    self._input_length, communicator=self.comm
                )
        assert self._output_handle is not None
        self._output_handle.set_data(qp_dstn, partial=True)
        if self.config.output_mode != "return":
            self._output_handle.write_chunk(start, end)
        return qp_dstn



import os

import ceci
import numpy as np
import pytest

from rail.core.stage import RailPipeline, RailStage
from rail.tools.table_tools import ColumnMapper, TableConverter
from rail.utils.path_utils import RAILDIR


def test_pipeline() -> None:
    DS = RailStage.data_store
    DS.__class__.allow_overwrite = True
    DS.clear()

    input_file = os.path.join(
        RAILDIR, "rail/examples_data/goldenspike_data/data//test_flow_data.pq"
    )
    bands = ["u", "g", "r", "i", "z", "y"]
    _band_dict = {band: f"mag_{band}_lsst" for band in bands}
    rename_dict = {f"mag_{band}_lsst": f"{band}_lsst" for band in bands}
    _post_grid = [float(x) for x in np.linspace(0.0, 5, 21)]

    col_remapper_test = ColumnMapper.make_stage(
        name="col_remapper_test", hdf5_groupname="", columns=rename_dict
    )

    table_conv_test = TableConverter.make_stage(
        name="table_conv_test", output_format="numpyDict", seed=12345
    )

    pipe = ceci.Pipeline.interactive()
    stages = [
        col_remapper_test,
        table_conv_test,
    ]
    for stage in stages:
        pipe.add_stage(stage)

    table_conv_test.connect_input(col_remapper_test)

    pipe.initialize(
        dict(input=input_file), dict(output_dir=".", log_dir=".", resume=False), None
    )

    pipe.save("stage.yaml")

    pr = ceci.Pipeline.read("stage.yaml")
    pr.run()

    os.remove("stage.yaml")
    os.remove("stage_config.yml")

    outputs = pr.find_all_outputs()
    for output_ in outputs.values():
        try:
            os.remove(output_)
        except FileNotFoundError:
            pass
    logfiles = [f"{stage.instance_name}.out" for stage in pr.stages]
    for logfile_ in logfiles:
        try:
            os.remove(logfile_)
        except FileNotFoundError:
            pass


def test_golden_v2() -> None:
    DS = RailStage.data_store
    DS.__class__.allow_overwrite = True
    DS.clear()
    pipe = RailPipeline()

    input_file = os.path.join(
        RAILDIR, "rail/examples_data/goldenspike_data/data//test_flow_data.pq"
    )
    bands = ["u", "g", "r", "i", "z", "y"]
    _band_dict = {band: f"mag_{band}_lsst" for band in bands}
    rename_dict = {f"mag_{band}_lsst": f"{band}_lsst" for band in bands}
    _post_grid = [float(x) for x in np.linspace(0.0, 5, 21)]

    pipe.col_remapper_test = ColumnMapper.build(
        hdf5_groupname="",
        columns=rename_dict,
    )

    pipe.table_conv_test = TableConverter.build(
        connections=dict(
            input=pipe.col_remapper_test.io.output  # pylint: disable=no-member
        ),
        output_format="numpyDict",
        seed=12345,
    )

    pipe.initialize(
        dict(input=input_file), dict(output_dir=".", log_dir=".", resume=False), None
    )
    pipe.save("stage.yaml")

    pr = ceci.Pipeline.read("stage.yaml")
    pr.run()

    os.remove("stage.yaml")
    os.remove("stage_config.yml")

    outputs = pr.find_all_outputs()
    for output_ in outputs.values():
        try:
            os.remove(output_)
        except FileNotFoundError:
            pass
    logfiles = [f"{stage.instance_name}.out" for stage in pr.stages]
    for logfile_ in logfiles:
        try:
            os.remove(logfile_)
        except FileNotFoundError:
            pass


def test_load_pipeline() -> None:
    train_z_class = RailPipeline.load_pipeline_class(
        "rail.pipelines.estimation.train_z_pipeline.TrainZPipeline"
    )

    check = RailPipeline.get_pipeline_class("TrainZPipeline")
    assert check == train_z_class

    RailPipeline.print_classes()

    with pytest.raises(KeyError):
        RailPipeline.get_pipeline_class("Does not exist")



import os
from typing import Any

import numpy as np
import pandas as pd
import pytest

from rail.core.data import DATA_STORE, DataHandle, TableHandle
from rail.creation.degraders.addRandom import AddColumnOfRandom
from rail.creation.degraders.quantityCut import QuantityCut


@pytest.fixture
def data() -> DataHandle:
    """Some dummy data to use below."""

    DS = DATA_STORE()
    DS.__class__.allow_overwrite = True

    # generate random normal data
    rng = np.random.default_rng(0)
    x = rng.normal(loc=26, scale=1, size=(100, 7))

    # replace redshifts with reasonable values
    x[:, 0] = np.linspace(0, 2, x.shape[0])

    # return data in handle wrapping a pandas DataFrame
    df = pd.DataFrame(x, columns=["redshift", "u", "g", "r", "i", "z", "y"])
    return DS.add_data("data", df, TableHandle, path="dummy.pq")


@pytest.fixture
def data_forspec() -> DataHandle:
    """Some dummy data to use below."""

    DS = DATA_STORE()
    DS.__class__.allow_overwrite = True

    # generate random normal data
    rng = np.random.default_rng(0)
    x = rng.normal(loc=26, scale=1, size=(200000, 7))

    # replace redshifts with reasonable values
    x[:, 0] = np.linspace(0, 2, x.shape[0])

    # return data in handle wrapping a pandas DataFrame
    df = pd.DataFrame(x, columns=["redshift", "u", "g", "r", "i", "z", "y"])
    return DS.add_data("data_forspec", df, TableHandle, path="dummy_forspec.pq")


@pytest.mark.parametrize(
    "cuts,error",
    [
        (1, TypeError),
        ({"u": "cut"}, TypeError),
        ({"u": dict()}, TypeError),
        ({"u": [1, 2, 3]}, ValueError),
        ({"u": [1, "max"]}, TypeError),
        ({"u": [2, 1]}, ValueError),
        ({"u": TypeError}, TypeError),
    ],
)
def test_QuantityCut_bad_params(cuts: Any, error: type[Exception]) -> None:
    """Test bad parameters that should return Type and Value errors"""
    with pytest.raises(error):
        QuantityCut.make_stage(cuts=cuts)


def test_QuantityCut_returns_correct_shape(  # pylint: disable=redefined-outer-name
    data: Any,
) -> None:
    """Make sure QuantityCut is returning the correct shape"""

    cuts = {
        "u": 30,
        "redshift": (1, 2),
    }
    degrader = QuantityCut.make_stage(cuts=cuts)

    degraded_data = degrader(data).data

    assert (
        degraded_data.shape
        == data.data.query("u < 30 & redshift > 1 & redshift < 2").shape
    )
    os.remove(degrader.get_output(degrader.get_aliased_tag("output"), final_name=True))

    degrader_w_flag = QuantityCut.make_stage(
        name="degrader_w_flag", cuts=cuts, drop_rows=False
    )

    degraded_data_w_flag = degrader_w_flag(data).data

    test_mask = np.zeros(len(data.data), dtype=int)
    out_indices = data.data.query("u < 30 & redshift > 1 & redshift < 2").index.values
    test_mask[out_indices] = 1

    assert (degraded_data_w_flag["flag"] == test_mask).all()
    os.remove(
        degrader_w_flag.get_output(
            degrader_w_flag.get_aliased_tag("output"), final_name=True
        )
    )


def test_add_random(data: Any) -> None:  # pylint: disable=redefined-outer-name
    add_random = AddColumnOfRandom.make_stage()

    test_data = add_random(data, seed=1234).data
    assert len(test_data[add_random.config.col_name]) == len(data.data)

    os.remove(
        add_random.get_output(add_random.get_aliased_tag("output"), final_name=True)
    )



"""
Abstract base classes for Informers.
These superstages ingest prior information, including training sets and explicit
priors, and prepare a model that can be used to produce photo-z data products.
They are distinguished by their input data types, and the models they output can
be used for their corresponding Estimator, Summarizer, or Classifier stages.
"""

from typing import Any, Generator
import numpy as np

import qp

from rail.core.common_params import SHARED_PARAMS
from rail.core.data import (DataHandle, ModelHandle, QPHandle, TableHandle,
                            TableLike)
from rail.core.stage import RailStage
import tables_io


class CatInformer(RailStage):
    """The base class for informing models used to make photo-z data products
    from catalog-like inputs (i.e., tables with fluxes in photometric bands among
    the set of columns).

    Estimators use a generic "model", the details of which depends on the sub-class.
    Most estimators will have associated Informer classes, which can be used to inform
    those models.

    (Note, "Inform" is more generic than "Train" as it also applies to algorithms that
    are template-based rather than machine learning-based.)

    Informer will produce as output a generic "model", the details of which depends on the sub-class.

    They take as "input" catalog-like tabular data, which is used to "inform" the model.
    """

    name = "CatInformer"
    config_options = RailStage.config_options.copy()
    inputs = [("input", TableHandle)]
    outputs = [("model", ModelHandle)]
    config_options.update(
        hdf5_groupname=SHARED_PARAMS,
    )

    def __init__(self, args: Any, **kwargs: Any) -> None:
        """Initialize Informer that can inform models for redshift estimation"""
        super().__init__(args, **kwargs)
        self.model = None

    def inform(self, training_data: TableLike) -> DataHandle:
        """The main interface method for Informers

        This will attach the input_data to this `Informer`
        (for introspection and provenance tracking).

        Then it will call the run(), validate() and finalize() methods, which need to
        be implemented by the sub-classes.

        The run() method will need to register the model that it creates to this Estimator
        by using `self.add_data('model', model)`.

        Finally, this will return a ModelHandle providing access to the trained model.

        Parameters
        ----------
        input_data
            dictionary of all input data, or a `TableHandle` providing access to it

        Returns
        -------
        ModelHandle
            Handle providing access to trained model
        """

        self.set_data("input", training_data)
        self.validate()
        self.run()
        self.finalize()
        return self.get_handle("model")

    def _convert_table_format(self, data: TableLike, out_fmt_str: str="numpyDict") -> TableLike: # pragma: no cover
        """
        Utility function to convert existing Tabular data to a numpy dictionary,
        ingestable for most informer and estimators.
        To be called in run().
        """
        # required format for informer/estimator
        out_fmt = tables_io.types.TABULAR_FORMAT_NAMES[out_fmt_str] 
        out_data = tables_io.convert(data, out_fmt)
        # overwrite set_data
        return out_data
        

class PzInformer(RailStage):
    """The base class for informing models used to make photo-z data products from
    existing ensembles of p(z) distributions.

    PzInformer can use a generic "model", the details of which depends on the sub-class.
    Some summarizer will have associated PzInformer classes, which can be used to inform
    those models.

    (Note, "Inform" is more generic than "Train" as it also applies to algorithms that
    are template-based rather than machine learning-based.)

    PzInformer will produce as output a generic "model", the details of which depends on the sub-class.

    They take as "input" a qp.Ensemble of per-galaxy p(z) data, which is used to "inform" the model.
    """

    name = "PzInformer"
    config_options = RailStage.config_options.copy()
    config_options.update(
        hdf5_groupname=SHARED_PARAMS,
        chunk_size=SHARED_PARAMS,
    )
        
    inputs = [("input", QPHandle), ("truth", TableHandle)]
    outputs = [("model", ModelHandle)]

    def __init__(self, args: Any, **kwargs: Any) -> None:
        """Initialize Informer that can inform models for redshift estimation"""
        super().__init__(args, **kwargs)
        self.model = None
        self.model_handle: ModelHandle | None = None

    def _setup_iterator(self) -> Generator:

        itrs = []
        input_itr = self.input_iterator("input", groupname="")
        truth_itr = self.input_iterator("truth", groupname=self.config.hdf5_groupname)

        if input_itr:
            itrs.append(input_itr)
        else:  # pragma: no cover
            return []
        if truth_itr:  # pragma: no cover
            itrs.append(truth_itr)
            
        for it in zip(*itrs):
            first = True
            for s, e, d in it:
                true_redshift: np.ndarray | None = None
                if first:
                    start = s
                    end = e
                    qp_ens = d
                    first = False
                else:  # pragma: no cover
                    true_redshift = d[self.config.redshift_col]

            yield start, end, qp_ens, true_redshift

    def inform(
        self, training_data: qp.Ensemble | str="None", truth_data: TableLike | str = "None",
    ) -> dict[str, DataHandle]:
        """The main interface method for Informers

        This will attach the input_data to this `Informer`
        (for introspection and provenance tracking).

        Then it will call the run(), validate() and finalize() methods, which need to
        be implemented by the sub-classes.

        The run() method will need to register the model that it creates to this Estimator
        by using `self.add_data('model', model)`.

        Finally, this will return a ModelHandle providing access to the trained model.

        Parameters
        ----------
        input_data
            Per-galaxy p(z), and any ancilary data associated with it

        truth_data
            Table with the true redshifts

        Returns
        -------
        dict[str, DataHandle]
            Handle providing access to trained model
        """
        self.set_data("input", training_data)
        self.set_data("truth", truth_data)
        self.validate()
        self.run()
        self.finalize()
        self._model_handle = self.get_handle("model")
        return dict(
            model=self._model_handle,
        )

    def run(self) -> None:
        iterator = self._setup_iterator()
        first = True
        self._initialize_run()
        self._output_handle = None
        for s, e, test_data, truth_data in iterator:
            print(f"Process {self.rank} running estimator on chunk {s:,} - {e:,}")
            self._process_chunk(s, e, test_data, truth_data, first)
            first = False
        self._finalize_run()

    def _initialize_run(self) -> None:
        self._model_handle = None

    def _finalize_run(self) -> None:
        assert self.model is not None
        self.add_data("model", self.model)
        self._model_handle = self.get_handle("model")        
        assert self._model_handle is not None        
        self._model_handle.write()

    def _process_chunk(
        self, start: int, end: int, data: qp.Ensemble, truth_data: np.ndarray, first: bool
    ) -> None:
        return
        



"""
A classifier that uses pz point estimate to assign
tomographic bins with uniform binning. 
"""

import numpy as np
from ceci.config import StageParameter as Param

from rail.core.data import Hdf5Handle
from rail.estimation.classifier import PZClassifier


class EqualCountClassifier(PZClassifier):
    """Classifier that simply assign tomographic
    bins based on point estimate according to SRD"""

    name = "EqualCountClassifier"
    config_options = PZClassifier.config_options.copy()
    config_options.update(
        id_name=Param(
            str,
            "",
            msg="Column name for the object ID in the input data, if empty the row index is used as the ID.",
        ),
        point_estimate=Param(str, "zmode", msg="Which point estimate to use"),
        zmin=Param(float, 0.0, msg="Minimum redshift of the sample"),
        zmax=Param(float, 3.0, msg="Maximum redshift of the sample"),
        nbins=Param(int, 5, msg="Number of tomographic bins"),
        no_assign=Param(int, -99, msg="Value for no assignment flag"),
    )
    outputs = [("output", Hdf5Handle)]

    def run(self) -> None:
        test_data = self.get_data("input")
        npdf = test_data.npdf

        try:
            zb = np.squeeze(test_data.ancil[self.config.point_estimate])
        except KeyError as msg:
            raise KeyError(
                f"{self.config.point_estimate} is not contained in the data ancil, "
                "you will need to compute it explicitly."
            ) from msg

        # tomographic bins with equal number density
        sortind = np.argsort(zb)
        cum = np.arange(1, (len(zb) + 1))
        bin_index = np.zeros(len(zb))
        for ii in range(self.config.nbins):
            perc1 = ii / self.config.nbins
            perc2 = (ii + 1) / self.config.nbins
            ind = (cum / cum[-1] > perc1) & (cum / cum[-1] <= perc2)
            useind = sortind[ind]
            bin_index[useind] = int(ii + 1)

        if self.config.id_name != "":
            # below is commented out and replaced by a redundant line
            # because the data doesn't have ID yet
            # obj_id = test_data[self.config.id_name]
            obj_id = np.arange(npdf)
        elif self.config.id_name == "":
            # ID set to row index
            obj_id = np.arange(npdf)
            self.config.id_name = "row_index"

        class_id = {self.config.id_name: obj_id, "class_id": bin_index}
        self.add_data("output", class_id)



import os
import pprint
from typing import Any

import ceci
import click
import yaml

from rail.cli.rail import options, scripts
from rail.core import __version__
from rail.interfaces.pz_factory import PZFactory
from rail.interfaces.tool_factory import ToolFactory
from rail.utils import catalog_utils


@click.group()
@click.version_option(__version__)
def cli() -> None:
    """RAIL utility scripts"""

    
@cli.command(name="render-nb")
@options.outdir(default="docs")
@options.clear_output()
@options.dry_run()
@options.inputs()
@options.skip()
def render_nb(
    outdir: str,
    clear_output: bool,
    dry_run: bool,
    inputs: list[str],
    skip: list[str],
    **_kwargs: Any,
) -> int:
    """Render jupyter notebooks"""
    return scripts.render_nb(outdir, clear_output, dry_run, inputs, skip)


@cli.command(name="get-data")
@options.bpz_demo_data()
@options.verbose_download()
def get_data(verbose: bool, **kwargs: Any) -> int:  # pragma: no cover
    """Downloads data from NERSC (if not already found)"""
    scripts.get_data(verbose, **kwargs)
    return 0


@cli.command(name="clone-source", deprecated="In the future use 'rail dev clone-source ...' instead")
@options.outdir(default="..")
@options.git_mode()
@options.dry_run()
@options.package_file()
def clone_source(
    outdir: str,
    git_mode: options.GitMode,
    dry_run: bool,
    package_file: str,
    **_kwargs: Any,
) -> int:
    """Install packages from source"""
    scripts.clone_source(outdir, git_mode, dry_run, package_file)
    return 0


@cli.command(name="update-source", deprecated="In the future use 'rail dev update-source ...' instead")
@options.outdir(default="..")
@options.dry_run()
@options.package_file()
def update_source(outdir: str, dry_run: bool, package_file: str, **_kwargs: Any) -> int:
    """Update packages from source"""
    scripts.update_source(outdir, dry_run, package_file)
    return 0


@cli.command(name="install", deprecated="In the future use 'rail dev install' ... instead")
@options.outdir(default="..")
@options.dry_run()
@options.from_source()
@options.package_file()
def install(
    outdir: str, dry_run: bool, from_source: bool, package_file: str, **_kwargs: Any
) -> int:
    """Install rail packages one by one, to be fault tolerant"""
    scripts.install(outdir, from_source, dry_run, package_file)
    return 0


@cli.group(name="dev")
def dev_group() -> None:
    """Development related sub-commands"""


@dev_group.command(name="clone-source")
@options.outdir(default="..")
@options.git_mode()
@options.dry_run()
@options.package_file()
def clone_source(
    outdir: str,
    git_mode: options.GitMode,
    dry_run: bool,
    package_file: str,
    **_kwargs: Any,
) -> int:
    """Install packages from source"""
    scripts.clone_source(outdir, git_mode, dry_run, package_file)
    return 0


@dev_group.command(name="update-source")
@options.outdir(default="..")
@options.dry_run()
@options.package_file()
def update_source(outdir: str, dry_run: bool, package_file: str, **_kwargs: Any) -> int:
    """Update packages from source"""
    scripts.update_source(outdir, dry_run, package_file)
    return 0


@dev_group.command(name="git-status")
@options.outdir(default="..")
@options.dry_run()
@options.package_file()
def git_status(outdir: str, dry_run: bool, package_file: str, **_kwargs: Any) -> int:
    """Run git status on all packages"""
    scripts.git_status(outdir, dry_run, package_file)
    return 0


@dev_group.command(name="git-diff")
@options.outdir(default="..")
@options.dry_run()
@options.package_file()
def git_diff(outdir: str, dry_run: bool, package_file: str, **_kwargs: Any) -> int:
    """Run git diff on all packages"""
    scripts.git_diff(outdir, dry_run, package_file)
    return 0


@dev_group.command(name="git-describe")
@options.outdir(default="..")
@options.dry_run()
@options.package_file()
def git_describe(outdir: str, dry_run: bool, package_file: str, **_kwargs: Any) -> int:
    """Run git describe --tags on all packages"""
    scripts.git_describe(outdir, dry_run, package_file)
    return 0


@dev_group.command(name="install")
@options.outdir(default="..")
@options.dry_run()
@options.from_source()
@options.package_file()
def install(
    outdir: str, dry_run: bool, from_source: bool, package_file: str, **_kwargs: Any
) -> int:
    """Install rail packages one by one, to be fault tolerant"""
    scripts.install(outdir, from_source, dry_run, package_file)
    return 0


@cli.group(name="info")
def info_group() -> None:
    """Print information about the rail ecosystem"""


@info_group.command(name="all")
@options.outdir(default="..")
def print_all(**kwargs: Any) -> int:
    """Print everything"""
    scripts.info(print_all=True, **kwargs)
    return 0


@info_group.command(name="packages")
@options.outdir(default="..")
def print_packages(**kwargs: Any) -> int:
    """Print package related infor"""
    scripts.info(print_packages=True, **kwargs)
    return 0


@info_group.command(name="namespaces")
@options.outdir(default="..")
def print_namespaces(**kwargs: Any) -> int:
    """Print package related infor"""
    scripts.info(print_namespaces=True, **kwargs)
    return 0


@info_group.command(name="modules")
@options.outdir(default="..")
def print_modules(**kwargs: Any) -> int:
    """Print package related infor"""
    scripts.info(print_modules=True, **kwargs)
    return 0


@info_group.command(name="tree")
@options.outdir(default="..")
def print_tree(**kwargs: Any) -> int:
    """Print package related infor"""
    scripts.info(print_tree=True, **kwargs)
    return 0


@info_group.command(name="stages")
@options.outdir(default="..")
def print_stages(**kwargs: Any) -> int:
    """Print package related infor"""
    scripts.info(print_stages=True, **kwargs)
    return 0


@cli.group(name="pipe")
def pipe_group() -> None:
    """Pipeline related sub-commands"""

    
@pipe_group.command(name="build")
@options.pipeline_class()
@options.output_yaml()
@options.catalog_tag()
@options.stages_config()
@options.outdir()
@options.inputs()
def build_pipe(
    pipeline_class: str,
    output_yaml: str,
    catalog_tag: str,
    stages_config: dict,
    outdir: str,
    inputs: dict[str, str],
) -> int:  # pragma: no cover
    """Build a pipeline yaml file"""
    input_dict = {}
    for input_ in inputs:
        tokens = input_.split("=")
        assert len(tokens) == 2
        input_dict[tokens[0]] = tokens[1]
    scripts.build_pipeline(
        pipeline_class, output_yaml, catalog_tag, input_dict, stages_config, outdir
    )
    return 0

    
@pipe_group.command(name="run-stage")
@options.pipeline_yaml()
@options.stage_name()
@options.dry_run()
@options.inputs()
def run_stage(
    pipeline_yaml: str, stage_name: str, dry_run: bool, inputs: dict[str, str]
) -> int:  # pragma: no cover
    """Run a pipeline stage"""
    pipe = ceci.Pipeline.read(pipeline_yaml)
    input_dict = {}
    for input_ in inputs:
        tokens = input_.split("=")
        assert len(tokens) == 2
        input_dict[tokens[0]] = tokens[1]
    com = pipe.generate_stage_command(stage_name, **input_dict)
    if dry_run:
        print(com)
    else:
        os.system(com)
    return 0


@cli.group(name="run")
def run_group() -> None:
    """Sub-commands to run particular stages"""


@run_group.command(name="estimator")
@options.stage_name()
@options.stage_class()
@options.stage_module()
@options.stages_config()
@options.model_file()
@options.catalog_tag()
@options.dry_run()
@options.input_file()
@options.params()
def run_estimator(
    stage_name: str,
    stage_class: str,
    stage_module: str,
    stages_config: str | None,
    model_file: str,
    catalog_tag: str,
    dry_run: bool,
    input_file: str,
    params: dict,
) -> int:  # pragma: no cover
    """Run a RAIL pz estimation stage"""
    if catalog_tag:
        catalog_utils.apply_defaults(catalog_tag)

    if stages_config not in [None, "none", "None"]:
        assert stages_config is not None
        with open(stages_config, "r", encoding="utf-8") as fin:
            config_data = yaml.safe_load(fin)
            if stage_name in config_data:
                kwargs = config_data[stage_name]
            elif isinstance(config_data, dict):
                kwargs = config_data
            else:
                raise ValueError(
                    f"Config file {stages_config} is not properly constructed"
                )
    else:
        kwargs = {}

    kwargs.update(**options.args_to_dict(params))

    stage = PZFactory.build_cat_estimator_stage(
        stage_name=stage_name,
        class_name=stage_class,
        module_name=stage_module,
        model_path=model_file,
        data_path="dummy.in",
        **kwargs,
    )

    if dry_run:
        print(f"Running stage {stage.name} of type {type(stage)} on {input_file}")
        print("Stage config:   ")
        print_dict = stage.config.to_dict().copy()
        pprint.pprint(print_dict)
        return 0

    _output = PZFactory.run_cat_estimator_stage(
        stage,
        data_path=input_file,
    )
    return 0


@run_group.command(name="tool")
@options.stage_name()
@options.stage_class()
@options.stage_module()
@options.dry_run()
@options.input_file()
def run_tool(
    stage_name: str, stage_class: str, stage_module: str, dry_run: bool, input_file: str
) -> int:  # pragma: no cover
    """Run a RAIL Tool stage"""
    stage = ToolFactory.build_tool_stage(
        stage_name=stage_name,
        class_name=stage_class,
        module_name=stage_module,
        data_path="dummy.in",
    )

    if dry_run:
        print(f"Running stage {stage.name} of type {type(stage)} on {input_file}")
        print("Stage config:   ")
        print_dict = stage.config.to_dict().copy()
        pprint.pprint(print_dict)
        return 0

    _output = ToolFactory.run_tool_stage(
        stage,
        data_path=input_file,
    )

    return 0



""" Parameters that are shared between stages """

from typing import Any

from ceci.config import StageConfig
from ceci.config import StageParameter as Param

lsst_bands = "ugrizy"
lsst_mag_cols = [f"mag_{band}_lsst" for band in lsst_bands]
lsst_mag_err_cols = [f"mag_err_{band}_lsst" for band in lsst_bands]
lsst_def_err_dict: dict[str, str|None] = dict(zip(lsst_mag_cols, lsst_mag_err_cols))
lsst_def_err_dict["redshift"] = None

lsst_def_maglims = dict(
    mag_u_lsst=27.79,
    mag_g_lsst=29.04,
    mag_r_lsst=29.06,
    mag_i_lsst=28.62,
    mag_z_lsst=27.98,
    mag_y_lsst=27.05,
)
# default reddening parameters for LSST
lsst_def_a_env = dict(
    mag_u_lsst=4.81,
    mag_g_lsst=3.64,
    mag_r_lsst=2.70,
    mag_i_lsst=2.06,
    mag_z_lsst=1.58,
    mag_y_lsst=1.31,
)
lsst_err_band_replace = [0.1, 0.1, 0.1, 0.1, 0.1, 0.1]
lsst_zp_errors = [0.1, 0.1, 0.1, 0.1, 0.1, 0.1]
lsst_filter_list = [
    "DC2LSST_u",
    "DC2LSST_g",
    "DC2LSST_r",
    "DC2LSST_i",
    "DC2LSST_z",
    "DC2LSST_y",
]


SHARED_PARAMS = StageConfig(
    hdf5_groupname=Param(
        str, "photometry", msg="name of hdf5 group for data, if None, then set to ''"
    ),
    chunk_size=Param(
        int, 10000, msg="Number of object per chunk for parallel processing"
    ),
    zmin=Param(float, 0.0, msg="The minimum redshift of the z grid"),
    zmax=Param(float, 3.0, msg="The maximum redshift of the z grid"),
    nzbins=Param(int, 301, msg="The number of gridpoints in the z grid"),
    dz=Param(float, 0.01, msg="delta z in grid"),
    nondetect_val=Param(
        float, 99.0, msg="value to be replaced with magnitude limit for non detects"
    ),
    nonobserved_val=Param(
        float, -99.0, msg="guard value for non-observations"
    ),
    bands=Param(
        list, lsst_mag_cols, msg="Names of columns for magnitgude by filter band"
    ),
    err_bands=Param(
        list,
        lsst_mag_err_cols,
        msg="Names of columns for magnitgude errors by filter band",
    ),
    err_dict=Param(
        dict,
        lsst_def_err_dict,
        msg="dictionary that contains the columns that will be used to"
        "predict as the keys and the errors associated with that column as the values."
        "If a column does not havea an associated error its value shoule be `None`",
    ),
    mag_limits=Param(dict, lsst_def_maglims, msg="Limiting magnitdues by filter"),
    band_a_env=Param(dict, lsst_def_a_env, msg="Redenning parameters"),
    ref_band=Param(str, "mag_i_lsst", msg="band to use in addition to colors"),
    redshift_col=Param(str, "redshift", msg="name of redshift column"),
    id_col=Param(str, "object_id", msg="name of the object ID column"),
    object_id_col=Param(str, "objectId", msg="name of object id column"),
    zp_errors=Param(
        dtype=list,
        default=lsst_zp_errors,
        msg="BPZ adds these values in quadrature to the photometric errors",
    ),
    calc_summary_stats=Param(
        dtype=bool,
        default=False,
        msg="Compute summary statistics",
    ),
    calculated_point_estimates=Param(
        dtype=list,
        default=[],
        msg="List of strings defining which point estimates to automatically calculate using `qp.Ensemble`."
        "Options include, 'mean', 'mode', 'median'.",
    ),
    recompute_point_estimates=Param(
        dtype=bool,
        default=False,
        msg="Force recomputation of point estimates",
    ),
    replace_error_vals=Param(
        dtype=list,
        default=lsst_err_band_replace,
        msg="list of values to replace negative and nan mag err values",
    ),
    filter_list=Param(
        dtype=list,
        default=lsst_filter_list,
        msg="list of filter files names (with no '.sed' suffix). Filters must be"
        "in FILTER dir.  MUST BE IN SAME ORDER as 'bands'",
    ),
)


class SharedParams:
    """Class to store parameters shared between many stages"""

    try:
        _config_text = SHARED_PARAMS.numpy_style_help_text()
        __doc__: str | None = f"\n\nParameters\n----------\n{_config_text}"
    except Exception:  # pragma: no cover
        pass

    @staticmethod
    def copy_param(param_name: str) -> Param:
        """Return a copy of one of the shared parameters

        Parameters
        ----------
        param_name
            Name of the parameter to copy

        Returns
        -------
        Param
            Copied parameter
        """
        return SHARED_PARAMS.get(param_name).copy()

    @staticmethod
    def set_param_default(param_name: str, default_value: Any) -> None:
        """Change the default value of one of the shared parameters

        Parameters
        ----------
        param_name
            Name of the parameter to copy

        default_value
            New default value
        """
        try:
            SHARED_PARAMS.get(param_name).set_default(default_value)
        except AttributeError as msg:  # pragma: no cover
            raise KeyError(
                f"No shared parameter {param_name} in SHARED_PARAMS"
            ) from msg

    @staticmethod
    def set_param_defaults(**kwargs: Any) -> None:  # pragma: no cover
        """Change the default value of several of the shared parameters

        Parameters
        ----------
        **kwargs
            Key, value pairs of parameter names and default values
        """
        for key, val in kwargs.items():
            set_param_default(key, val)


copy_param = SharedParams.copy_param

set_param_default = SharedParams.set_param_default

set_param_defaults = SharedParams.set_param_defaults



"""Abstract base class defining a selector.

The key feature here is make selection to either the photometric or spectroscopic catalog. 
Intended subclasses spectroscopic selection, probability selection on a grid for the photometry, 
or pure photometric selection. 
"""

from ceci.config import StageParameter as Param

from rail.core.data import DataHandle, PqHandle, TableLike
from rail.core.stage import RailStage


class Selector(RailStage):
    """Base class Selector, which makes selection to the catalog

    Selector take "input" data in the form of pandas dataframes in Parquet
    files and provide as "output" another pandas dataframes written to Parquet
    files.
    """

    name = "Selector"
    config_options = RailStage.config_options.copy()
    config_options.update(
        drop_rows=Param(bool, True, msg="Drop selected rows from output table"),
        seed=Param(
            default=None,
            required=False,
            msg="Set to an `int` to force reproducible results.",
        ),
    )
    inputs = [("input", PqHandle)]
    outputs = [("output", PqHandle)]

    def __call__(self, sample: TableLike) -> DataHandle:
        """The main interface method for ``Selector``.

        Adds noise to the input catalog

        This will attach the input to this `Selector`

        Then it will call the select() which add a flag column to the catalog. flag=1 means
        selected, 0 means dropped.

        If dropRows = True, the dropped rows will not be presented in the output catalog,
        otherwise, all rows will be presented.

        Finally, this will return a PqHandle providing access to that output
        data.

        Parameters
        ----------
        sample:
            The sample to be selected

        Returns
        -------
        DataHandle
            A handle giving access to a table with selected sample
        """
        self.set_data("input", sample)
        self.run()
        self.finalize()
        return self.get_handle("output")

    def run(self) -> None:
        data = self.get_data("input")
        selection_mask = self._select()
        if self.config["drop_rows"]:
            out_data = data[selection_mask.astype(bool)]
        else:
            out_data = data.copy()
            out_data.insert(0, "flag", selection_mask)
        self.add_data("output", out_data)

    def _select(self) -> TableLike:  # pragma: no cover
        raise NotImplementedError("Selector._select()")



import glob
import os

import pytest

from rail.cli.rail import scripts
from rail.cli.rail.options import GitMode, args_to_dict


def test_args_to_dict() -> None:
    args = ["x=3", "y=2"]
    as_dict = args_to_dict(args)
    assert as_dict["x"] == "3"
    assert as_dict["y"] == "2"

    with pytest.raises(ValueError):
        args_to_dict(["x"])
    with pytest.raises(ValueError):
        args_to_dict(["x=3=3"])


def test_render_nb() -> None:
    nb_dir = "./tests/cli/"
    nb_files = glob.glob(os.path.join(nb_dir, "*.ipynb"))
    scripts.render_nb("docs", False, True, nb_files, skip=[])
    scripts.render_nb(
        "docs", True, True, nb_files, skip=["./tests/cli/single_number.ipynb"]
    )
    scripts.render_nb(
        "docs", True, False, nb_files, skip=["./tests/cli/single_number.ipynb"]
    )


def test_clone_source() -> None:
    scripts.clone_source("..", GitMode.ssh, True, "rail_packages.yml")
    scripts.clone_source("..", GitMode.https, True, "rail_packages.yml")
    scripts.clone_source("..", GitMode.cli, True, "rail_packages.yml")


def test_update_source() -> None:
    scripts.update_source("..", True, "rail_packages.yml")


def test_install() -> None:
    scripts.install("..", False, True, "rail_packages.yml")
    scripts.install("..", True, True, "rail_packages.yml")


def test_info() -> None:
    scripts.info(print_all=True)

    
def test_git_stuff() -> None:
    scripts.git_status("..", True, "rail_packages.yml")
    scripts.git_diff("..", True, "rail_packages.yml")
    scripts.git_describe("..", True, "rail_packages.yml")

    
def test_build_pipeline() -> None:
    scripts.build_pipeline(
        "rail.pipelines.estimation.train_z_pipeline.TrainZPipeline",
        "trainz_pipe.yaml",
        input_dict=dict(input="dummy.ibn"),
    )

    os.unlink("trainz_pipe.yaml")
    os.unlink("trainz_pipe_config.yml")



""" Base class for PipelineStages in Rail """

from __future__ import annotations

import os
from math import ceil
from typing import Any, Iterable, TypeVar

from ceci.config import StageParameter as Param
from ceci.pipeline import MiniPipeline
from ceci.stage import PipelineStage

from .data import DATA_STORE, DataHandle, DataLike, ModelHandle

T = TypeVar("T", bound="RailPipeline")
S = TypeVar("S", bound="RailStage")


class StageIO:
    """A small utility class for Stage Input/ Output

    This make it possible to get access to stage inputs and outputs
    as attributes rather that by using the get_handle() method.

    In short it maps

    a_stage.get_handle('input', allow_missing=True) to a_stage.input

    This allows users to be more concise when writing pipelines.
    """

    def __init__(self, parent: RailStage):
        self._parent = parent

    def __getattr__(self, item: str) -> DataHandle:
        return self._parent.get_handle(item, allow_missing=True)


class RailStageBuild:
    """A small utility class that building stages

    This provides a mechasim to get the name of the stage from the
    attribute name in the Pipeline the stage belongs to.

    I.e., we can do:

    a_pipe.stage_name = StageClass.build(...)

    And get a stage named 'stage_name', rather than having to do:

    a_stage = StageClass.make_stage(..)
    a_pipe.add_stage(a_stage)
    """

    def __init__(self, stage_class: type[RailStage], **kwargs: Any):
        self.stage_class: type[PipelineStage] = stage_class
        self._kwargs: dict = kwargs
        self._stage: RailStage | None = None

    @property
    def io(self) -> StageIO | None:  # pragma: no cover
        if self._stage:
            return self._stage.io
        return None

    def build(self, name: str) -> RailStage:
        """Actually build the stage, this is called by the pipeline the stage
        belongs to

        Parameters
        ----------
        name
            The name for this stage we are building

        Returns
        -------
        RailStage
            The newly built stage
        """
        self._stage = self.stage_class.make_and_connect(name=name, **self._kwargs)
        assert self._stage is not None
        return self._stage


class RailPipeline(MiniPipeline):
    """A pipeline intended for interactive use

    Mainly this allows for more concise pipeline specification, along the lines of:

    self.stage_1 = Stage1Class.build(...)
    self.stage_2 = Stage2Class.build(connections=dict(input=self.stage1.io.output), ...)

    And end up with a fully specified pipeline.
    """

    pipeline_classes: dict[str, type[RailPipeline]] = {}

    def __init_subclass__(cls) -> None:
        cls.pipeline_classes[cls.__name__] = cls

    @classmethod
    def print_classes(cls) -> None:
        for key, val in cls.pipeline_classes.items():
            print(f"{key} {val}")

    @classmethod
    def get_pipeline_class(cls, name: str) -> type[RailPipeline]:
        try:
            return cls.pipeline_classes[name]
        except KeyError as msg:
            raise KeyError(
                f"Could not find pipeline class {name} in {list(cls.pipeline_classes.keys())}"
            ) from msg

    @staticmethod
    def load_pipeline_class(class_name: str) -> type[RailPipeline]:
        """Import a particular RailPipeline subclass by name

        Parameters
        ----------
        class_name
            Full name of the class, e.g., rail.core.stage.RailPipeline

        Returns
        -------
        type[RailPipeline]
            Requested Pipeline sub-class
        """
        tokens = class_name.split(".")
        module = ".".join(tokens[:-1])
        class_name = tokens[-1]
        __import__(module)
        pipe_class = RailPipeline.get_pipeline_class(class_name)
        return pipe_class

    @staticmethod
    def build_and_write(
        class_name: str,
        output_yaml: str,
        input_dict: dict | None = None,
        stages_config: dict | None = None,
        output_dir: str = ".",
        log_dir: str = ".",
        **kwargs: Any,
    ) -> None:
        """Build a RailPipeline and write the config yaml for for it

        Parameters
        ----------
        class_name
            Full name of the class, e.g., rail.core.stage.RailPipeline

        output_yaml
            Path to the output yaml file

        input_dict
            Dict of all the inputs needed to run the pipeline

        stages_config
            Stage configuration overrides

        output_dir
            Directory to write pipeline outputs to

        log_dir
            Directory to write pipeline log files to

        **kwargs
            Passed as arguements to the pipeline constructor
        """
        pipe_class = RailPipeline.get_pipeline_class(class_name)
        pipe = pipe_class(**kwargs)

        full_input_dict = pipe_class.default_input_dict.copy()
        if input_dict is not None:
            full_input_dict.update(**input_dict)
        pipe.initialize(
            full_input_dict,
            dict(
                output_dir=output_dir,
                log_dir=log_dir,
                resume=False,
            ),
            stages_config,
        )
        pipe.save(output_yaml)

    def __init__(self) -> None:
        MiniPipeline.__init__(self, [], dict(name="mini"))

    def __setattr__(
        self, name: str, value: RailStageBuild | Any
    ) -> PipelineStage | Any:
        if isinstance(value, RailStageBuild):
            stage = value.build(name)
            self.add_stage(stage)
            return stage
        return MiniPipeline.__setattr__(self, name, value)


class RailStage(PipelineStage):
    """Base class for rail stages

    This inherits from `ceci.PipelineStage` and implements rail-specific data handling
    In particular, this provides some very useful features:

    1.  Access to the `DataStore`, which keeps track of the various data used in a pipeline, and
    provides access to each by a unique key.

    2.  Functionality to help manage multiple instances of a particular class of stage.
    The original ceci design didn't have a mechanism to handle this.  If you tried
    you would run into name clashes between the different instances.  In `ceci` 1.7 we
    added functionality to `ceci` to allow you to have multiple instances of a single class,
    in particular we distinguish between the class name (`cls.name`) and and the name of
    the particular instance (`self.instance_name`) and added aliasing for inputs and outputs,
    so that different instances of `PipelineStage` would be able to give different names
    to their inputs and outputs.  However, using that functionality in a consistent way
    requires a bit of care.  So here we are providing methods to do that, and to do it in
    a way that uses the `DataStore` to keep track of the various data products.

    Notes
    -----
    These methods typically take a tag as input (i.e., something like "input"),
    but use the "aliased_tag" (i.e., something like "inform_pz_input") when interacting
    with the DataStore.

    In particular, the `get_handle()`, `get_data()` and `input_iterator()` will get the data
    from the DataStore under the aliased tag.  E.g., if you call `self.get_data('input')` for
    a `Stage` that has aliased "input" to "special_pz_input", it will
    get the data associated to "special_pz_input" in the DataStore.

    Similarly, `add_handle()` and `set_data()` will add the data to the DataStore under the aliased tag
    e.g., if you call `self.set_data('input')` for a `Stage` that has
    aliased "input" to "special_pz_input", it will store the data in the DataStore
    under the key "special_pz_input".

    And `connect_input()` will do the alias lookup both on the input and output.
    I.e., it is the same as calling
    `self.set_data(inputTag, other.get_handle(outputTag, allow_missing=True), do_read=False)`
    """

    config_options = dict(
        output_mode=Param(str, "default", msg="What to do with the outputs")
    )

    data_store = DATA_STORE()

    def __init__(self, args: Any, **kwargs: Any) -> None:
        """Constructor:
        Do RailStage specific initialization"""
        super().__init__(args, **kwargs)
        self._input_length: int | None = None
        self.io = StageIO(self)
        self.stage_columns: list[str] | None = None

    @classmethod
    def make_and_connect(cls: type[S], **kwargs: Any) -> S:
        """Make a stage and connects it to other stages

        Notes
        -----
        kwargs are used to set stage configuration,
        the should be key, value pairs, where the key
        is the parameter name and the value is value we want to assign

        The 'connections' keyword is special, it is a dict[str, DataHandle]
        and should define the Input connections for this stage

        Returns
        -------
        A stage
        """
        connections = kwargs.pop("connections", {})
        stage = cls.make_stage(**kwargs)
        for key, val in connections.items():
            stage.set_data(key, val, do_read=False)
        return stage

    @classmethod
    def build(cls, **kwargs: Any) -> RailStageBuild:
        """Return an object that can be used to build a stage"""
        return RailStageBuild(cls, **kwargs)

    def get_handle(
        self, tag: str, path: str | None = None, allow_missing: bool = False
    ) -> DataHandle:
        """Gets a DataHandle associated to a particular tag

        Parameters
        ----------
        tag
            The tag (from cls.inputs or cls.outputs) for this data

        path
            The path to the data, only needed if we might need to read the data

        allow_missing
            If False this will raise a key error if the tag is not in the DataStore

        Returns
        -------
        DataHandle
            The handle that give access to the associated data
        """
        aliased_tag = self.get_aliased_tag(tag)
        handle = self.data_store.get(aliased_tag)
        if handle is None:
            if not allow_missing:
                raise KeyError(
                    f"{self.instance_name} failed to get data by handle {aliased_tag}, associated to {tag}"
                )
            handle = self.add_handle(tag, path=path)
        return handle

    def add_handle(
        self, tag: str, data: DataLike = None, path: str | None = None
    ) -> DataHandle:
        """Adds a DataHandle associated to a particular tag

        Parameters
        ----------
        tag
            The tag (from cls.inputs or cls.outputs) for this data

        data
            If not None these data will be associated to the handle

        path
            If not None, this will be the path used to read the data

        Returns
        -------
        DataHandle
            The handle that gives access to the associated data
        """
        aliased_tag = self.get_aliased_tag(tag)
        if aliased_tag in self._inputs:
            if path is None:
                path = self.get_input(aliased_tag)
            handle_type = self.get_input_type(tag)
        else:
            if path is None:
                path = self.get_output(aliased_tag)
            handle_type = self.get_output_type(tag)
        handle = handle_type(
            aliased_tag, path=path, data=data, creator=self.instance_name
        )
        print(
            f"Inserting handle into data store.  {aliased_tag}: {handle.path}, {handle.creator}"
        )
        self.data_store[aliased_tag] = handle
        return handle

    def get_data(self, tag: str, allow_missing: bool = True) -> DataLike:
        """Gets the data associated to a particular tag

        Notes
        -----
        1. This gets the data via the DataHandle, and can and will read the data
        from disk if needed.

        Parameters
        ----------
        tag
            The tag (from cls.inputs or cls.outputs) for this data

        allow_missing
            If False this will raise a key error if the tag is not in the DataStore

        Returns
        -------
        DataLike
            The data accesed by the handle assocated to the tag
        """
        handle = self.get_handle(tag, allow_missing=allow_missing)
        if not handle.has_data:
            handle.read()
        return handle()

    def set_data(
        self, tag: str, data: DataLike, path: str | None = None, do_read: bool = True
    ) -> DataLike:
        """Sets the data associated to a particular tag

        Notes
        -----
        1. If data is a DataHandle and tag is one of the input tags,
        then this will add an alias between the two, i.e., it will
        set `self.config.alias[tag] = data.tag`.  This allows the user to
        make connections between stages simply by passing DataHandles between
        them.

        Parameters
        ----------
        tag
            The tag (from cls.inputs or cls.outputs) for this data

        data
            The data being set,

        path
            Can be used to set the path for the data

        do_read
            If True, will read the data if it is not set

        Returns
        -------
        DataLike
            The data accesed by the handle assocated to the tag
        """
        # First we grab the handle that we will be using
        handle = self.get_handle(tag, path=path, allow_missing=True)

        if isinstance(data, DataHandle):
            # If we were passed a DataHandle, we use that
            aliased_tag = data.tag
            if tag in self.input_tags():
                self._aliases[tag] = aliased_tag
                if data.has_path:
                    self._inputs[tag] = data.path
                    handle.path = data.path
                if data.has_data:
                    handle.data = data.data
                elif do_read:
                    handle.read()
        elif not isinstance(data, (type(None), str)):
            # If we were passed data, we use that and reset the path
            handle.data = data
            if path in ['None', 'none', None]:
                handle.path = 'None'
        else:
            # Data is None, we use the path
            if path not in ['None', 'none', None]:
                # Path exists, use that                
                if not os.path.isfile(path):
                    raise FileNotFoundError(f"Unable to find file: {path}")
                handle.path = path
                if do_read:
                    # Read the data using the handle
                    handle.read()

        return handle.data

    def add_data(self, tag: str, data: DataLike = None) -> DataLike:
        """Adds a handle to the DataStore associated to a particular tag and
        attaches data to it.

        Parameters
        ----------
        tag
            The tag (from cls.inputs or cls.outputs) for this data

        data
            Data being added

        Returns
        -------
        DataLike
            The data accesed by the handle assocated to the tag
        """
        handle = self.add_handle(tag, data=data)
        return handle.data

    def input_iterator(self, tag: str, **kwargs: Any) -> Iterable:
        """Iterate the input assocated to a particular tag

        Parameters
        ----------
        tag
            The tag (from cls.inputs or cls.outputs) for this data

        **kwargs
            These will be passed to the Handle's iterator method

        Returns
        -------
        Fixme

        """
        handle = self.get_handle(tag, allow_missing=True)

        try:
            groupname = kwargs.get("groupname", self.config.hdf5_groupname)
        except Exception:
            groupname = None

        chunk_size = kwargs.get("chunk_size", self.config.chunk_size)

        on_disk: bool= False
        in_memory: bool= False
        if handle.path not in [None, 'None', 'none']:
            on_disk = True
        if handle.data is not None:
            in_memory = True
        
        if on_disk:
            self._input_length = handle.size(groupname=groupname)

            total_chunks_needed = ceil(self._input_length / chunk_size)
            # If the number of process is larger than we need, we reduce chunk_size
            # so that all of the processes have some data to work with.
            if total_chunks_needed < self.size:  # pragma: no cover
                self.config.chunk_size = int(ceil(self._input_length / self.size))
                chunk_size = self.config.chunk_size
                print(
                    "Warning: You are reserving more processes than needed, reducing chunk size to",
                    chunk_size,
                    "to use all of the processes",
                )

            kwcopy = dict(
                groupname=groupname,
                chunk_size=chunk_size,
                rank=self.rank,
                parallel_size=self.size,
            )
            kwcopy.update(**kwargs)
            return handle.iterator(**kwcopy)

        # If data is in memory and not in a file, it means is small enough to process it
        # in a single chunk.
        elif in_memory:  # pragma: no cover
            if self.config.hdf5_groupname:
                test_data = self.get_data(tag)[self.config.hdf5_groupname]
                self._input_length = self.get_handle(tag).data_size(
                    groupname=self.config.hdf5_groupname
                )
            else:
                test_data = self.get_data(tag)
                self._input_length = self.get_handle(tag).data_size()
            s = 0
            iterator = [[s, self._input_length, test_data]]
            return iterator

        # Data is neither on disk or in memory, return empty list
        else:
            return []
        
    def connect_input(
        self,
        other: PipelineStage,
        inputTag: str | None = None,
        outputTag: str | None = None,
    ) -> DataHandle:
        """Connect another stage to this stage as an input

        Parameters
        ----------
        other
             The stage whose output is being connected

        inputTag
             Which input tag of this stage to connect to.  None -> self.inputs[0]

        outputTag
             Which output tag of the other stage to connect to.  None -> other.outputs[0]

        Returns
        -------
        DataHandle
            The input handle for this stage
        """
        if inputTag is None:
            inputTag = self.inputs[0][0]  # pylint: disable=no-member
        if outputTag is None:
            outputTag = other.outputs[0][0]
        handle = other.get_handle(outputTag, allow_missing=True)
        return self.set_data(inputTag, handle, do_read=False)

    def _finalize_tag(self, tag: str) -> str:
        """Finalize the data for a particular tag.

        This can be overridden by sub-classes for more complicated behavior
        """
        handle = self.get_handle(tag, allow_missing=True)
        if self.config.output_mode == "default":
            assert handle.path is not None
            if not os.path.exists(handle.path) or not handle.partial:
                handle.write()
        final_name = PipelineStage._finalize_tag(self, tag)
        handle.path = final_name
        return final_name

    def _check_column_names(
        self, data: Any, columns_to_check: list[str], **kwargs: Any
    ) -> None:
        try:
            groupname = kwargs.get("groupname", self.config.hdf5_groupname)
        except Exception:  # pragma: no cover
            groupname = None

        if not groupname:
            groupname = None

        if isinstance(data, DataHandle) and not data.has_data:
            if data.has_path:
                # data handle only has a path, read the columns from the path
                path = data.path
                assert path is not None
                data._check_data_columns(
                    path, columns_to_check, parent_groupname=groupname, **kwargs
                )
            elif not data.has_path:  # pragma: no cover
                print("The data handle does not contain data or path.")

        else:
            # data has been read in, access the columns in the table/dictionary directly
            if isinstance(data, DataHandle) and data.has_data:
                if groupname in [None, ""]:
                    col_list = list(data.data.keys())
                else:
                    col_list = list(data.data[groupname].keys())
            else:
                # data is passed as a table
                if groupname in [None, ""]:
                    col_list = list(data.keys())
                else:
                    col_list = list(data[groupname].keys())
            # check columns
            intersection = set(columns_to_check).intersection(col_list)
            if len(intersection) < len(columns_to_check):
                diff = set(columns_to_check) - intersection
                raise KeyError("The following columns are not found: ", diff)

    def _get_stage_columns(self) -> None:
        self.stage_columns = None  # pragma: no cover

    def open_model(self, tag: str = "model", **kwargs: Any) -> ModelLike:
        """Load the mode and/or attach it to this Stage

        Parameters
        ----------
        tag
            Input tag associated to the model

        **kwargs
            Should include 'model', see notes

        Notes
        -----
        The keyword arguement 'model' should be either

        1. an object with a trained model,
        2. a path pointing to a file that can be read to obtain the trained model,
        3. or a `ModelHandle` providing access to the trained model.

        Returns
        -------
        Any
            The object encapsulating the trained model.
        """
        if tag not in self.input_tags():  # pragma: no cover
            raise KeyError(f"Stage {self} can not open model with input tag {tag}")
        model = kwargs.get(tag, None)
        if model is None or model == "None":  # pragma: no cover
            self.model = None
            return self.model
        if isinstance(model, str):
            self.model = self.set_data(tag, data=None, path=model)
            self.config[tag] = model
            return self.model
        if isinstance(model, ModelHandle):
            if model.has_path:
                self.config[tag] = model.path
        self.model = self.set_data(tag, model)
        return self.model



""" Utility functions """

import os

import rail
import rail.core

RAILDIR = os.path.abspath(os.path.join(os.path.dirname(rail.core.__file__), "..", ".."))


def find_rail_file(relpath: str) -> str:
    """Find a file somewhere in rail by searching the namespace path

    This lets us avoid issues that the paths can be different depending
    on if we have installed things from source or not
    """
    for path_ in rail.__path__:
        fullpath = os.path.abspath(os.path.join(path_, relpath))
        if os.path.exists(fullpath):
            return fullpath
    raise ValueError(f"Could not file {relpath} in {rail.__path__}")



"""Abstract base class defining a noisifier.

The key feature here is the run adds noise to the catalog. 
Intended subclasses are noisifier that adds LSST noise / other telescope noise
"""

from ceci.config import StageParameter as Param

from rail.core.data import DataHandle, PqHandle, TableLike
from rail.core.stage import RailStage


class Noisifier(RailStage):
    """Base class Noisifier, which adds noise to the input catalog

    Noisifier take "input" data in the form of pandas dataframes in Parquet
    files and provide as "output" another pandas dataframes written to Parquet
    files.
    """

    name = "Noisifier"
    config_options = RailStage.config_options.copy()
    config_options.update(
        seed=Param(
            default=None,
            required=False,
            msg="Set to an `int` to force reproducible results.",
        ),
    )
    inputs = [("input", PqHandle)]
    outputs = [("output", PqHandle)]

    def _initNoiseModel(self) -> None:  # pragma: no cover
        raise NotImplementedError("Noisifier._initNoiseModel()")

    def _addNoise(self) -> None:  # pragma: no cover
        raise NotImplementedError("Noisifier._addNoise()")

    def __call__(self, sample: TableLike, seed: int | None = None) -> DataHandle:
        """The main interface method for ``Noisifier``.

        Adds noise to the input catalog

        This will attach the input to this `Noisifier`

        Then it will call the _initNoiseModel() and _addNoise(), which need to be
        implemented by the sub-classes.

        The _initNoiseModel() method will initialize the noise model of the sub-classes, and
        store the noise model as self.noiseModel

        The _addNoise() method will add noise to the flux and magnitude of the column of the
        catalog.

        The finalize() method will check the end results (like preserving number of rows)

        Finally, this will return a PqHandle providing access to that output
        data.

        Parameters
        ----------
        sample :
            The sample to be degraded
        seed :
            An integer to set the numpy random seed

        Returns
        -------
        output_data : PqHandle
            A handle giving access to a table with degraded sample
        """
        if seed is not None:
            self.config.seed = seed
        self.set_data("input", sample)

        self.run()
        self.finalize()
        return self.get_handle("output")

    def run(self) -> None:
        self._initNoiseModel()
        self._addNoise()



import rail.stages
from rail.core import RailStage
from rail.utils.testing_utils import check_stage_params

rail.stages.import_and_attach_all()


def test_all_stage_params() -> None:
    fail_list = []

    for _key, val in RailStage.pipeline_stages.items():
        fail_string = check_stage_params(val[0])
        if fail_string:
            fail_list.append(fail_string)

    if fail_list:
        for fail_ in fail_list:
            print(fail_)
        raise TypeError(f"Found {len(fail_list)} unparseable parameters")



from typing import Any

from ceci.config import StageParameter as Param
from qp.metrics.concrete_metric_classes import DistToDistMetric

from rail.core.data import QPHandle
from rail.evaluation.evaluator import Evaluator


class DistToDistEvaluator(Evaluator):
    """Evaluate the performance of a photo-z estimator against reference PDFs"""

    name = "DistToDistEvaluator"
    config_options = Evaluator.config_options.copy()
    config_options.update(
        limits=Param(
            list,
            [0.0, 3.0],
            required=False,
            msg="The default end points for calculating metrics on a grid.",
        ),
        dx=Param(
            float,
            0.01,
            required=False,
            msg="The default step size when calculating metrics on a grid.",
        ),
        num_samples=Param(
            int,
            100,
            required=False,
            msg="The number of random samples to select for certain metrics.",
        ),
    )
    inputs = [("input", QPHandle), ("truth", QPHandle)]

    metric_base_class = DistToDistMetric

    def _process_chunk(self, data_tuple: Any, first: bool) -> None:
        start = data_tuple[0]
        end = data_tuple[1]
        estimate_data = data_tuple[2]
        reference_data = data_tuple[3]

        self._process_all_chunk_metrics(
            estimate_data, reference_data, start, end, first
        )

    def _process_all(self, data_tuple: Any) -> None:
        estimate_data = data_tuple[0]
        reference_data = data_tuple[1]

        self._process_all_metrics(estimate_data, reference_data)
